"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


class DLCLiveError(Exception):
    """ Generic error type for incorrect use of the DLCLive class """

    pass


class DLCLiveWarning(Warning):
    """ Generic warning for incorrect use of the DLCLive class """

    pass
